import numpy as np
import xarray as xr

from redplanet.Mag.sh.loader import get_dataset

from redplanet.helper_functions.docstrings.main import substitute_docstrings



@substitute_docstrings
def get(
    lon       : float | np.ndarray,
    lat       : float | np.ndarray,
    quantity  : str  = 'total',
    as_xarray : bool = False
) -> float | np.ndarray | xr.DataArray:
    """
    Get magnetic field values at the specified coordinates. Dataset must be loaded first, see `redplanet.Mag.sh.load(...)`.

    Parameters
    ----------
    {param.lon}
    {param.lat}
    quantity : str, optional
        Options are: ['radial', 'theta', 'phi', 'total', 'potential'], by default 'total'.
    {param.as_xarray}

    Returns
    -------
    {return.GriddedData}

        Units are nanotesla [nT].

    Raises
    ------
    ValueError
        If `quantity` is not one of ['radial', 'theta', 'phi', 'total', 'potential.
    """

    q = ['radial', 'theta', 'phi', 'total', 'potential']
    if quantity not in q:
        raise ValueError(f"Quantity {quantity} is not in list of supported quantities: {q}.")

    dat_mag = get_dataset()

    return dat_mag.get_values(
        lon = lon,
        lat = lat,
        var = quantity,
        as_xarray = as_xarray,
    )



from pathlib import Path
import zipfile

import numpy as np
import pandas as pd

from redplanet.DatasetManager.main import _get_fpath_dataset
from redplanet.helper_functions.coordinates import _plon2slon
from redplanet.helper_functions.docstrings.main import substitute_docstrings





_dat_depths: pd.DataFrame | None = None

@substitute_docstrings
def get_dataset(as_dict: bool = False) -> pd.DataFrame:
    """
    Get the full magnetic source depth dataset.

    Data (0.038 MiB) is provided by {@Gong2021_data.n}. The full paper discusses/analyzes the models in detail ({@Gong2021_paper.p}).


    Parameters
    ----------
    as_dict : bool, optional
        If True, return the data as a list of dictionaries. Default is False.


    Returns
    -------
    pd.DataFrame | list[dict]
        Information about all 412 dipoles. Columns are:

        - `lon` : float
            - Longitude in range [-180, 180].
        - `lat` : float
            - Latitude in range [-90, 90].
        - `chi_reduced` : float
            - "reduced chi^2 value of the best fitting model"
        - `cap_radius_km` : list[float]
            - "angular radii of the magnetized caps (best-fit, and 1-sigma lower/upper limits)"
        - `depth_km` : list[float]
            - "magnetization depth (best-fit, and 1-sigma lower/upper limits)"
        - `dipole_mment_Am2` : list[float]
            - "square root of the metric N<M^2>V^2 [in A m^2] (best-fit, and 1-sigma lower/upper limits)"

        Note that the 1-sigma lower/upper values are NaN when the minimum reduced chi^2 value of the best fitting model is outside the 1-sigma confidence level of the reduced chi^2 that were obtained from Monte Carlo simulations.
    """
    if _dat_depths is None:
        _load()
    if as_dict:
        return _dat_depths.to_dict(orient='records')
    return _dat_depths





@substitute_docstrings
def _load() -> None:
    """
    Load the magnetic source depth dataset.

    For more information on the dataset, see `help(redplanet.Mag.depth.get_nearest)`.

    {note._load}
    """

    fname2level = {
        '20_17_8_134_150.dat'      : 'middle',
        '20_17_8_134_150_lower.dat': 'lower',
        '20_17_8_134_150_upper.dat': 'upper',
    }
    cols = ['lat', 'lon', 'cap_radius_km', 'depth_km', 'dipole_moment_Am2', 'chi2_reduced']
    dict_dfs = {}

    fpath_zip = _get_fpath_dataset('Gong & Weiczorek, 2021')

    ## open zipfile and iterate over files
    with zipfile.ZipFile(fpath_zip, 'r') as zipped_dir:
        for zipped_file in zipped_dir.infolist():

            fpath_dat = zipped_file.filename

            level = fname2level.get(Path(fpath_dat).name)  ## this will be one of: ('middle', 'lower', 'upper', None)
            if level is None:
                continue

            with zipped_dir.open(fpath_dat) as fpath_dat_unzipped:
                df = pd.read_csv(
                    fpath_dat_unzipped,
                    sep = r'\s+',
                    names = cols
                )

            df['lon'] = df['lon'].apply(_plon2slon)
            df.replace( {-1e100: np.nan}, inplace=True )

            dict_dfs[level] = df


    ## directly copy constant columns
    cols_const = ['lon', 'lat', 'chi2_reduced']
    global _dat_depths
    _dat_depths = dict_dfs['middle'][cols_const].copy()

    ## for "leveled" quantities, merge them into a numpy array ordered `[middle, lower, upper]`
    cols_merge = ['cap_radius_km', 'depth_km', 'dipole_moment_Am2']
    for col in cols_merge:
        list_arrays = []
        for level in ['middle', 'lower', 'upper']:
            list_arrays.append( dict_dfs[level][col].to_numpy() )

        # _dat_depths[col] = np.column_stack(list_arrays).tolist()
        _dat_depths[col] = list( np.column_stack(list_arrays) )

    return



import numpy as np
import cartopy.geodesic as cg


'''
Mars reference ellipsoid (Oblate ellipsoid), 2009
- Parameters:
    - Semimajor axis   : 3395428 m
    - Flattening       : 0.005227617843759314
    - GM               : 42828372000000.0 m³/s²
    - Angular velocity : 7.0882181e-05 rad/s
- Sources:
    - Main:
        - Ardalan, A. A., Karimi, R., & Grafarend, E. W. (2009). A New Reference Equipotential Surface, and Reference Ellipsoid for the Planet Mars. Earth, Moon, and Planets, 106, 1-13.
        - https://doi.org/10.1007/s11038-009-9342-7
    - (Found it here:)
        - https://www.fatiando.org/boule/latest/ellipsoids.html
'''
_semimajor_m = 3395428
_flattening = 0.005227617843759314





'''
TODO: implement this myself at some point since I'm not using cartopy for anything else
    (for now we're just directly calling cartopy, but I wanted to put it in a module right away so it's easier to change later)

https://scitools.org.uk/cartopy/docs/latest/reference/generated/cartopy.geodesic.Geodesic.html
'''

__mars_geodesic: cg.Geodesic = None

def __get_mars_geodesic() -> cg.Geodesic:
    global __mars_geodesic
    if __mars_geodesic is None:
        __mars_geodesic = cg.Geodesic(
            radius     = _semimajor_m,
            flattening = _flattening,
        )
    return __mars_geodesic



def get_distance(
    start : list | np.ndarray,
    end   : list | np.ndarray,
) -> np.ndarray:
    """
    Calculate the geodesic distance between two points on the surface of Mars.

    Parameters
    ----------
    start : list | np.ndarray
        Array of shape (2) or (n_points, 2) containing the longitude and latitude coordinates of the starting point(s).
    end : list | np.ndarray
        Similar to `start`, but for the ending point(s).

    Returns
    -------
    np.ndarray
        Array of shape (n_points) [? verify this] containing the geodesic distances between the corresponding starting and ending point(s).

    See Also
    --------
    [cartopy.geodesic.Geodesic.circle](https://scitools.org.uk/cartopy/docs/latest/reference/generated/cartopy.geodesic.Geodesic.html)

    Notes
    -----
    Details on the Mars reference ellipsoid (oblate ellipsoid) used for these calculations:

    - Parameters:
        - Semimajor axis   : 3395428 m
        - Flattening       : 0.005227617843759314
        - GM               : 42828372000000.0 m³/s²
        - Angular velocity : 7.0882181e-05 rad/s
    - Sources:
        - Main:
            - Ardalan, A. A., Karimi, R., & Grafarend, E. W. (2009). A New Reference Equipotential Surface, and Reference Ellipsoid for the Planet Mars. Earth, Moon, and Planets, 106, 1-13.
            - https://doi.org/10.1007/s11038-009-9342-7
        - (Found it here:)
            - https://www.fatiando.org/boule/latest/ellipsoids.html

    For geodesic calculations, we use [cartopy.geodesic.Geodesic.circle](https://scitools.org.uk/cartopy/docs/latest/reference/generated/cartopy.geodesic.Geodesic.html).
    """
    geodesic = __get_mars_geodesic()
    return geodesic.inverse(start, end)



def move_forward(
    start    : list | np.ndarray,
    azimuth  : float | list | np.ndarray,
    distance : float,
) -> np.ndarray:
    """
    Calculate the coordinates of a point on the surface of Mars after moving a certain geodesic distance along a given angle.

    Parameters
    ----------
    start : list | np.ndarray
        Array of shape (2) containing the longitude and latitude coordinates of the starting point.
    azimuth : float | list | np.ndarray
        Array of shape (n_points) containing the azimuth angle(s) to "move forward" in degrees, where 0 is north and 90 is east.
    distance : float
        The geodesic distance to move forward in meters.

    Returns
    -------
    np.ndarray
        Array of shape (2) or (n_points, 2) [? verify this] containing the longitude and latitude coordinates of the point(s) after moving forward.

    See Also
    --------
    For more details about the reference ellipsoid and geodesic calculations, see `redplanet.helper_functions.geodesy.get_distance`.
    """
    geodesic = __get_mars_geodesic()
    return geodesic.direct(start, azimuth, distance)[:,:2]



def make_circle(
    lon       : float,
    lat       : float,
    radius    : float,  ## TODO: should this be suffixed with '_m' or '_km' to indicate units...?
    n_samples : int  = 180,
    endpoint  : bool = False,
) -> np.ndarray:
    """
    Generate a geodesic circle of points on the surface of Mars.

    Parameters
    ----------
    lon : float
        Longitude coordinate of the center of the circle, in range [-180, 360].
    lat : float
        Latitude coordinate of the center of the circle, in range [-90, 90].
    radius : float
        Radius of the circle in meters.
    endpoint : bool, optional
        If True, include the starting point at the end of the circle. Default is False.

    Returns
    -------
    np.ndarray
        Array of shape (n_samples, 2) [? verify this] containing the longitude and latitude coordinates of the circle points.

    See Also
    --------
    For more details about the reference ellipsoid, see `redplanet.helper_functions.geodesy.get_distance`.
    """
    geodesic = __get_mars_geodesic()
    return geodesic.circle(lon, lat, radius, n_samples, endpoint)



import numpy as np
import xarray as xr

from redplanet.GRS.loader import get_dataset

from redplanet.helper_functions.docstrings.main import substitute_docstrings



@substitute_docstrings
def get(
    element   : str,
    lon       : float | np.ndarray,
    lat       : float | np.ndarray,
    quantity  : str  = 'concentration',
    normalize : bool = False,
    as_xarray : bool = False
) -> float | np.ndarray | xr.DataArray:
    """
    Get GRS element concentration/sigma values at the specified coordinates.

    Data (0.41 MiB) is provided by {@GRS_data.n}. The full paper discusses/analyzes the models in detail ({@GRS_paper.p}).


    Parameters
    ----------
    element : str
        Element name. Options are: ['al', 'ca', 'cl', 'fe', 'h2o', 'k', 'si', 's', 'th'].
    {param.lon}
    {param.lat}
    quantity : str, optional
        Return 'concentration' or 'sigma' values, by default 'concentration'.
    normalize : bool, optional
        If True, normalize the element quantity to a volatile-free (Cl, H2O, S) basis, by default False.

        > "The GRS instrument measures elemental abundances in the top-most tens of centimeters of the Martian surface, and thus is strongly influenced by near-surface soils, ice and dust deposits. These sediments broadly represent the bulk chemistry of the Martian upper crust when renormalized to a volatile-free basis [Taylor and McLennan, 2009] and as such, K and Th values must be renormalized to a H2O-, S-, and Cl-free basis to better reflect bulk crustal values." ({@Hahn2011.p})
    {param.as_xarray}


    Returns
    -------
    {return.GriddedData}

        Units are mass fraction (out of one).


    Raises
    ------
    ValueError
        - `element` is not one of ['al', 'ca', 'cl', 'fe', 'h2o', 'k', 'si', 's', 'th'].
        - `quantity` is not one of ['concentration', 'sigma'].
        - `normalize` is True and `element` is one of ['cl', 'h2o', 's'] -- you can't normalize a volatile element to a volatile-free basis!
    """

    ## input validation
    e = ['al','ca','cl','fe','h2o','k','si','s','th']
    v = ['cl','h2o','s']
    q = ['concentration','sigma']

    if element not in e:
        raise ValueError(f"Element {element} is not in list of supported elements: {e}.")

    if quantity not in q:
        raise ValueError(f"Quantity {quantity} is not in list of supported quantities: {q}.")

    if normalize and (element in v):
        raise ValueError(f"Can't normalize a volatile element ('{element}') to a volatile-free (cl, h2o, s) basis.")


    ## get data & normalize
    dat_grs = get_dataset()

    dat = dat_grs.get_values(
        lon = lon,
        lat = lat,
        var = f'{element}_{quantity}',
        as_xarray = as_xarray,
    )

    if normalize:
        volatiles = dat_grs.get_values(
            lon = lon,
            lat = lat,
            var = f'cl+h2o+s_{quantity}',
            as_xarray = as_xarray,
        )
        dat = dat / (1 - volatiles)

    return dat



import numpy as np

from redplanet.DatasetManager.main import _get_fpath_dataset
from redplanet.helper_functions.GriddedData import GriddedData

from redplanet.helper_functions.docstrings.main import substitute_docstrings





_dat_topo: GriddedData | None = None

@substitute_docstrings
def get_dataset() -> GriddedData:
    """
    {fulldoc.get_dataset_GriddedData}
    """
    if _dat_topo is None:
        raise ValueError('Topography dataset not loaded. Use `redplanet.Crust.topo.load(<model_params>)`.')
    return _dat_topo

@substitute_docstrings
def get_metadata() -> dict:
    """
    {fulldoc.get_metadata}
    """
    return dict(get_dataset().metadata)





@substitute_docstrings
def load(model: str = None) -> None:
    """
    Load a topography model.

    Parameters
    ----------
    model : str
        Name of the topography model to load. Options are:

        - `'SH_10km'` (17 MB) — Spherical harmonic model of topography from {@MOLA_shcoeffs.n} evaluated to degree 1066, corresponding to a spatial resolution of 10 km.
        - `'SH_5km'` (70 MB) — Same as `'SH_10km'` but evaluated to degree 2133, corresponding to a spatial resolution of 5 km.
        - `'DEM_463m'` (2 GB) — MGS MOLA Digital Elevation Model 463m ({@DEM_463m.p}).
        - `'DEM_200m'` (11 GB) — MGS MOLA - MEX HRSC Blended Digital Elevation Model 200m ({@DEM_200m.p}).

        For description of our modifications to the original data, see notes section.

    Raises
    ------
    ValueError
        If an invalid model name is provided.

    Notes
    -----
    For 'SH_' options, we start with a spherical harmonic model of the shape of Mars based on MOLA data, provided by {@MOLA_shcoeffs.n}. Then we subtract the geoid height and geoid reference radius, which yields the planet's surface relief with respect to the geoid (i.e. topography). We assume a spherical harmonic model of degree $L$ has spatial resolution $\\frac{2 \\pi R_{\\text{Mars}}}{2L+1}$. Results are saved as a binary file, which can be loaded as a memory-mapped 2D numpy array of 16-bit integers for very fast access speeds without occupying RAM. For more information and our code, see ["datasets/Crust/topo/SH" in the GitHub repo.](https://github.com/Humboldt-Penguin/redplanet/tree/main/datasets/Crust/topo/SH){target="_blank"} -- TODO*

    For 'DEM_' options, we modify the original data files by reprojecting to the "Mars 2000 Sphere" model (radius = 3,396,190 km). We convert the original "TIFF" file format to a binary file, which can be loaded as a memory-mapped 2D numpy array of 16-bit integers for very fast access speeds without occupying RAM. For more information and our code, see ["datasets/Crust/topo/DEM" in the GitHub repo.](https://github.com/Humboldt-Penguin/redplanet/tree/main/datasets/Crust/topo/DEM){target="_blank"} -- TODO*

    * I'll eventually have a section on my website to describe datasets and how we modified them -- add a link to that here.
    """

    info = {
        'SH_10km': {
            'shape'    : (2135, 4269),
            'dtype'    : np.int16,
            'lon'      : np.linspace(0, 360, 4269),
            'lat'      : np.linspace(-90, 90, 2135),
            'metadata': {
                'title'   : 'Spherical harmonic models of the shape of Mars (MOLA) — 10 km resolution, corresponding to degree 1066',
                'units'   : 'm',
                'citation': 'Wieczorek, M. (2024). Spherical harmonic models of the shape of Mars (MOLA) [Dataset]. Zenodo. https://doi.org/10.5281/zenodo.10820719',
            },
        },
        'SH_5km': {
            'shape'    : (4269, 8537),
            'dtype'    : np.int16,
            'lon'      : np.linspace(0, 360, 8537),
            'lat'      : np.linspace(-90, 90, 4269),
            'metadata': {
                'title'   : 'Spherical harmonic models of the shape of Mars (MOLA) — 5 km resolution, corresponding to degree 2133',
                'units'   : 'm',
                'citation': 'Wieczorek, M. (2024). Spherical harmonic models of the shape of Mars (MOLA) [Dataset]. Zenodo. https://doi.org/10.5281/zenodo.10820719',
            },
        },
        'DEM_463m': {
            'shape'    : (23041, 46081),
            'dtype'    : np.int16,
            'nan_value': -99_999,  ## data is stored as int16 which doesn't support `np.nan`, so we use this sentinel value.
            'lon'      : -179.9960938347692 + 0.007812330461578525 * np.arange(46081),
            'lat'      : -89.99376946560506 + 0.00781206004494716  * np.arange(23041),
            'metadata': {
                'title'   : 'Mars MGS MOLA DEM 463m',
                'units'   : 'm',
                'citation': 'Goddard Space Flight Center (2003). Mars MGS MOLA DEM 463m [Dataset]. Astrogeology PDS Annex, U.S. Geological Survey. https://astrogeology.usgs.gov/search/map/mars_mgs_mola_dem_463m',
            },
        },
        'DEM_200m': {
            'shape'    : (53347, 106694),
            'dtype'    : np.int16,
            'nan_value': -99_999,  ## data is stored as int16 which doesn't support `np.nan`, so we use this sentinel value.
            'lon'      : -179.9983129395848 + 0.0033741208306410017 * np.arange(106694),
            'lat'      : -89.99753689179012 + 0.0033741208306410004 * np.arange(53347),
            'metadata': {
                'title'   : 'Mars MGS MOLA - MEX HRSC Blended DEM Global 200m',
                'units'   : 'm',
                'citation': 'Fergason, R. L., Hare, T. M., & Laura, J. (2018). Mars MGS MOLA - MEX HRSC Blended DEM Global 200m v2 [Dataset]. Astrogeology PDS Annex, U.S. Geological Survey. https://astrogeology.usgs.gov/search/map/mars_mgs_mola_mex_hrsc_blended_dem_global_200m',
            },
        },
    }

    if model not in info:
        raise ValueError(f"Invalid topography model: '{model}'. Options are: {list(info.keys())}.")


    # if model.startswith('DEM_'):

    fpath = _get_fpath_dataset(model)

    dat = np.memmap(
        fpath,
        mode  = 'r',
        dtype = info[model]['dtype'],
        shape = info[model]['shape'],
    )

    metadata = info[model]['metadata']
    metadata['fpath'] = fpath

    global _dat_topo
    _dat_topo = GriddedData(
        lon       = info[model]['lon'],
        lat       = info[model]['lat'],
        is_slon   = True,
        data_dict = {'topo': dat},
        metadata  = metadata,
    )


    # else:
    #     raise ValueError(f"THE DEVELOPER MESSED UP. THIS SHOULD NOT HAPPEN.")

    return



from pathlib import Path
from platformdirs import user_cache_dir    # type: ignore



_dirpath_datacache: Path = None
# _lock_thread = Lock() ## not sure if necessary... (use like: `with _lock_thread: ...`)



def get_dirpath_datacache() -> Path:
    """
    Get the data path where datasets are downloaded/cached. Initializes to default path ('/home/<user>/.cache/redplanet/') if not set.

    Returns
    -------
    Path

    Notes
    -----
    [DEVELOPER NOTES:]

    - It's good design to defer initialization of default value until first access in `get_dirpath_datacache()`...
        - This way, we avoid any overhead or side-effects of executing any potentially expensive / unnecessary code (esp related to file system operations) during the module import (i.e. lazy initialization).
    """
    ## Lazy load
    if _dirpath_datacache is None:
        set_dirpath_datacache(
            Path(user_cache_dir(appname='redplanet')).resolve()
        )
    return _dirpath_datacache



def set_dirpath_datacache(target_path: str | Path) -> None:
    """
    Set the data path where datasets will be downloaded/cached.

    Parameters
    ----------
    target_path : str | Path
        The file system path to store datasets.

    Raises
    ------
    TypeError
        Path must be a string or a Path object.
    ValueError
        Invalid path string provided.
    """
    ## Input type validation && conversion to Path object
    match target_path:

        case Path():
            target_path = target_path.resolve()

        case str():
            try:
                target_path = Path(target_path).resolve()
            except Exception as e:
                raise ValueError(f'Invalid path string provided: {target_path}\n{e}')

        case _:
            raise TypeError('Path must be a string or a Path object.')

    ## Proceed
    global _dirpath_datacache
    _dirpath_datacache = target_path
    return



## Define public API (self note: https://stackoverflow.com/questions/44834/what-does-all-mean-in-python)
__all__ = []



from redplanet import (
    ## meta package management
    user_config,

    ## load/access datasets
    Craters,
    Crust,
    GRS,
    Mag,

    ## high-level analysis
    analysis,
)

__all__.extend([
    'user_config',

    'Craters',
    'Crust',
    'GRS',
    'Mag',

    'analysis',
])
